"""
Plotting class for KIPET
"""
# Standard library imports
from pathlib import Path
import os
import time
import sys

# Thirdparty library imports 
import plotly.graph_objects as go
import plotly.io as pio

# Kipet library imports
"""
Constants used for plotting
"""
# Default Matlab colors
colors_rgb = [(0,    0.4470,    0.7410),
          (0.8500,    0.3250,    0.0980),
          (0.9290,    0.6940,    0.1250),
          (0.4940,    0.1840,    0.5560),
          (0.4660,    0.6740,    0.1880),
          (0.3010,    0.7450,    0.9330),
          (0.6350,    0.0780,    0.1840)
          ]

# Convert to rgb format used in plotly
colors = ['rgb(' + ','.join([str(int(255*c)) for c in color]) + ')' for color in colors_rgb]

colors += ['#4285F4', '#DB4437', '#F4B400', '#0F9D58',
                     '#185ABC', '#B31412', '#EA8600', '#137333',
                     '#d2e3fc', '#ceead6']

class PlotObject():
    
    """This will hold the relevant information needed to make a plot in KIPET"""
    
    def __init__(self, reaction_model=None, jupyter=False):
        
        self.reaction_model = reaction_model
        self.name = reaction_model.name
        self.color_num = 0
        self.filename = None
        self.jupyter = jupyter
        
    def _make_line_trace(self, fig, x, y, name, color):
        """Make line traces
        
        """
        line = dict(color=colors[color], width=4)
        fig.add_trace(
            go.Scatter(x=x,
                       y=y,
                       name=name,
                       line=line,
               )
            )
        return None
    
    def _make_marker_trace(self, fig, x, y, name, color, marker_options):
        """Make marker traces
        
        """
        fig.add_trace(
            go.Scatter(x=x,
                       y=y,
                       name=name,
                       mode='markers',
                       marker={**marker_options, 'color':colors[self.color_num]}),
                    )
        return None

    def _fig_finishing(self, fig, pred, plot_name='Z'):
        """Finish the plots before showing
        
        """
        x_data = [t for t in pred.index]
        x_axis_mod = 0.025*(float(x_data[-1]) - float(x_data[0]))
        fig.update_xaxes(range=[float(x_data[0])-x_axis_mod, float(x_data[-1])+x_axis_mod])
        fig.update_xaxes(zeroline=True, zerolinewidth=2, zerolinecolor='#4e4e4e')
        fig.update_yaxes(zeroline=True, zerolinewidth=2, zerolinecolor='#4e4e4e')

        if self.filename is None:
            t = time.localtime()
            date = f'{t.tm_year}-{t.tm_mon:02}-{t.tm_mday:02}-{t.tm_hour:02}-{t.tm_min:02}-{t.tm_sec:02}'
            # filename = f'{self.name}-{plot_name}-{date}.html'
            filename = f'{plot_name}.html'
            
        if self.jupyter:
            chart_dir = Path.cwd().joinpath('charts', f'{self.name}-{date}')
            plot_method = pio.show
            fig.update_layout(         
                autosize=False,
                width=1200,
                height=800,
                margin=dict(
                    l=50,
                    r=50,
                    b=50,
                    t=50,
                    pad=4
                    ),
                )
        else:
            calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
            chart_dir = Path(calling_file_name).joinpath('charts', f'{self.name}-{date}')
            plot_method = pio.write_html
        
        chart_dir.mkdir(parents=True, exist_ok=True)
        filename = chart_dir.joinpath(filename)
        if not self.jupyter:
           print(f'Plot saved as: {filename}')

        plot_method(fig, file=filename.as_posix(), auto_open=True)
    
        return None

    def _state_plot(self, fig, var, pred, exp, use_spectral_format=False):
        """Plot the state profiles
        
        """
        self._make_line_trace(fig=fig,
                              x=pred.index,
                              y=pred[var],
                              name=var,
                              color=self.color_num,
                              )
        marker_options = {'size': 15,
                           'opacity': 0.75,
                          }
        label = 'exp.'   
        if use_spectral_format:
            marker_options = {'size': 10,
                              'opacity': 0.5,
                             }
            label = 'spectral'         
        if exp is not None and var in exp.columns:
            self._make_marker_trace(fig=fig,
                                    x=exp.index,
                                    y=exp[var],
                                    name=f'{var} ({label})',
                                    color=self.color_num,
                                    marker_options=marker_options,
                                    )

    def _plot_all_Z(self):
        """Plot all concentration profiles
        
        """
        fig = go.Figure()
        use_spectral_format = False
        pred = getattr(self.reaction_model.results, 'Z')
        if hasattr(self.reaction_model.results, 'Cm'):
            exp = getattr(self.reaction_model.results, 'Cm')
        elif hasattr(self.reaction_model.results, 'C'):
            exp = getattr(self.reaction_model.results, 'C')
            use_spectral_format = True
        else:
            exp = None
        for i, col in enumerate(pred.columns):
            self._state_plot(fig, col, pred, exp, use_spectral_format=use_spectral_format)
            self.color_num += 1
        self.color_num = 0
        var_data = self.reaction_model.components[col]
        state = f'{self.reaction_model.components[col].state}'.capitalize()
        title = f'Model: {self.reaction_model.name} | Concentration Profiles'
        time_scale = f'Time [{self.reaction_model.ub.TIME_BASE}]'

        state_units = self._get_proper_unit_str(var_data)
        # state_units = var_data.units.u
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title=f'{state} [{state_units}]',
                )
        self._fig_finishing(fig, pred, plot_name='all-concentration-profiles')

    def _plot_Z(self, var):
        """Plot state profiles
        
        Args:
            var (str): concentration variable
        
        """
        fig = go.Figure()
        use_spectral_format = False
        pred = getattr(self.reaction_model.results, 'Z')
        if hasattr(self.reaction_model.results, 'Cm'):
            exp = getattr(self.reaction_model.results, 'Cm')
            self._state_plot(fig, var, pred, exp)
        elif hasattr(self.reaction_model.results, 'C'):
            exp = getattr(self.reaction_model.results, 'C')
            use_spectral_format = True
            self._state_plot(fig, var, pred, exp, use_spectral_format=use_spectral_format)
        else:
            exp = None
        var_data = self.reaction_model.components[var]
        state = f'{self.reaction_model.components[var].state}'.capitalize()
        description = f'| Description: {var_data.description}' if var_data.description is not None else ''
        title = f'Model: {self.reaction_model.name} | Variable: {var_data.name} {description}'
        time_scale = f'Time [{self.reaction_model.ub.TIME_BASE}]'
        # state_units = var_data.units.u
        state_units = self._get_proper_unit_str(var_data)
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title=f'{state} [{state_units}]',
                )
        self._fig_finishing(fig, pred, plot_name=f'{var}-concentration-profile')
        
    def _plot_Y(self, var, extra=None):
        """Plot state profiles

        Args:
            var (str): algebraic variable
        
        """
        fig = go.Figure()
        pred = self.reaction_model.results.Y
        if hasattr(self.reaction_model.results, 'UD'):
            exp = self.reaction_model.results.UD
        else:
            exp = None
        self._state_plot(fig, var, pred, exp)
        var_data = self.reaction_model.algebraics[var]
        description = f'| Description: {var_data.description}' if var_data.description is not None else ''
        title = f'Model: {self.reaction_model.name} | Variable: {var_data.name} {description}'
        time_scale = f'Time [{self.reaction_model.ub.TIME_BASE}]'
        # state_units = var_data.units
        state_units = self._get_proper_unit_str(var_data, check_expr=True)
        fig.update_layout(
            title=title,
            xaxis_title=f'{time_scale}',
            yaxis_title=f'[{state_units}]',
            )
        self._fig_finishing(fig, pred, plot_name=f'{var}-profile')

    def _plot_X(self, var):
        """Plot state profiles
        
        Args:
            var (str): state variable
        
        """
        fig = go.Figure()
        pred = getattr(self.reaction_model.results, 'X')
        if hasattr(self.reaction_model.results, 'U'):
            exp = getattr(self.reaction_model.results, 'U')
        else:
            exp = None
        self._state_plot(fig, var, pred, exp)
        var_data = self.reaction_model.states[var]
        description = f'| Description: {var_data.description}' if var_data.description is not None else ''
        title = f'Model: {self.reaction_model.name} | Variable: {var_data.name} {description}'
        time_scale = f'Time [{self.reaction_model.ub.TIME_BASE}]'
        state = f'{var_data.description}'.capitalize() if var_data.description is not None else 'State' 
        # state_units = var_data.units
        state_units = self._get_proper_unit_str(var_data)
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title=f'{state} [{state_units}]',
                )
        self._fig_finishing(fig, pred, plot_name=f'{var}-state-profile')
        
    def _plot_all_S(self):
        """Plot all S profiles
        
        """
        fig = go.Figure()
        pred = getattr(self.reaction_model.results, 'S')
        exp = None
        for i, col in enumerate(pred.columns):
            self._state_plot(fig, col, pred, exp)
            self.color_num += 1
        self.color_num = 0
        description = f'Description: Single species absorbance profiles'
        title = f'Model: {self.reaction_model.name} | {description}'
        time_scale = f'Wavelength [centimeter]'
        state = 'Absorbance'
        state_units = 'liter / mol / centimeter'
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title=f'{state} [{state_units}]',
                )
        self._fig_finishing(fig, pred, plot_name='all-absorbance-spectra')

    def _plot_S(self, var):
        """Plot individual S profile
        
        """
        fig = go.Figure()
        pred = getattr(self.reaction_model.results, 'S')
        exp = None
        self._state_plot(fig, var, pred, exp)
        var_data = self.reaction_model.results.S[var]
        description = f'| Description: Single species absorbance profiles'
        title = f'Model: {self.reaction_model.name} | Variable: {var} {description}'
        time_scale = f"Wavelength [centimeter]"
        state = 'Absorbance'
        state_units = 'liter / mol / centimeter'
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title=f'{state} [{state_units}]',
                )
        self._fig_finishing(fig, pred, plot_name=f'{var}-absorbance-spectra')
        
    def _plot_step(self, var, extra=None):
    
        fig = go.Figure()
        pred = self.reaction_model.results.step
        self._state_plot(fig, var, pred, None)
        title = f'Model: {self.reaction_model.name} | Variable: {var} | Step Function'
        time_scale = f'Time [{self.reaction_model.ub.TIME_BASE}]'
        fig.update_layout(
            title=title,
            xaxis_title=f'{time_scale}',
            yaxis_title=f'[dimensionless]',
            )
        self._fig_finishing(fig, pred, plot_name=f'{var}-step-profile')
        
    
    def _get_proper_unit_str(self, var_data, check_expr=False):
        
        try:
            str_units = var_data.units.u
        except:
            str_units = var_data.units
            
        if check_expr:
            str_units = str(self.reaction_model.alg_obj.exprs[var_data.name].units)
            
        return str_units
        


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kipet',
 'kipet.common',
 'kipet.core_methods',
 'kipet.dev_tools',
 'kipet.mixins',
 'kipet.model_components',
 'kipet.nsd_funs',
 'kipet.post_model_build',
 'kipet.top_level',
 'kipet.variance_methods',
 'kipet.visuals']

package_data = \
{'': ['*']}

install_requires = \
['Pint>=0.16.1,<0.17.0',
 'PyYAML>=5.4.1,<6.0.0',
 'Pyomo>=5.7.3,<6.0.0',
 'ipopt>=0.3.0,<0.4.0',
 'matplotlib>=3.3.4,<4.0.0',
 'numpy>=1.20.1,<2.0.0',
 'pandas>=1.2.2,<2.0.0',
 'plotly>=4.14.3,<5.0.0',
 'scipy>=1.6.0,<2.0.0']

setup_kwargs = {
    'name': 'kipet',
    'version': '0.1.53',
    'description': 'An all-in-one tool for fitting kinetic models using spectral and other state data',
    'long_description': '# <img alt="KIPET" src="branding/kipetlogo_full.svg" height="60">\n\n[![](https://img.shields.io/github/license/salvadorgarciamunoz/kipet)](https://github.com/salvadorgarciamunoz/kipet/blob/master/LICENSE)\n[![](https://img.shields.io/github/last-commit/salvadorgarciamunoz/kipet)](https://github.com/salvadorgarciamunoz/kipet/)\n[![](https://img.shields.io/pypi/wheel/kipet)](https://pypi.org/manage/project/kipet/release/0.1.1/)\n<br>\n\n[![](https://img.shields.io/badge/Install%20with-pip-green)]()\n[![](https://img.shields.io/pypi/v/kipet.svg?style=flat)](https://pypi.org/pypi/kipet/)\n<br>\n\n[![](https://anaconda.org/kwmcbride/kipet/badges/installer/conda.svg)]()\n[![Anaconda-Server Badge](https://anaconda.org/kwmcbride/kipet/badges/version.svg)](https://anaconda.org/kwmcbride/kipet)\n[![](https://anaconda.org/kwmcbride/kipet/badges/latest_release_date.svg)]()\n[![](https://anaconda.org/kwmcbride/kipet/badges/platforms.svg)]()\n\n\nKIPET is a Python package designed to simulate, and estimate parameters from \nchemical reaction systems through the use of maximum likelihood principles,\nlarge-scale nonlinear programming and discretization methods. \n\n- **Documentation:** - https://kipet.readthedocs.io\n- **Examples and Tutorials** - https://github.com/kwmcbride/kipet_examples\n- **Source code:** - https://github.com/salvadorgarciamunoz/kipet\n- **Bug reports:** - https://github.com/salvadorgarciamunoz/kipet/issues\n\nIt has the following functionality:\n\n - Simulate a reactive system described with DAEs\n - Solve the DAE system with collocation methods\n - Pre-process data\n - Estimate variances of noise from the model and measurements\n - Estimate kinetic parameters from spectra or concentration data across 1 or \n  multiple experiments with different conditions\n - Estimate confidence intervals of the estimated parameters\n - Able to estimate variances and parameters for problems where there is dosing / inputs into the system\n - Provide a set of tools for estimability analysis\n - Allows for wavelength selection of most informative wavelengths from a dataset\n - Visualize results\n\n\nInstallation\n------------\n\nA packaged version of KIPET can be installed using:\n\n    pip install kipet\n\nIf you run into errors when installing KIPET using pip, try installing the following packages beforehand:\n\n    pip install Cython numpy six\n    pip install kipet\n\nYou may also install KIPET with poetry (this method is recommended):\n\n    poetry add kipet\n\nFinally, if you are using Anaconda, KIPET can be installed using:\n\n    conda install -c kwmcbride kipet\n\nAdditionally, KIPET may be installed directly from the repository (if you want the latest version, simply install the desired branch (#branch)):\n\n    poetry add git+http://github.com/salvadorgarciamunoz/kipet#master\n\nNaturally you can simply clone or download the repository.\n\nLicense\n------------\n\nGPL-3\n\n\nAuthors\n----------\n\n    - Kevin McBride - Carnegie Mellon University\n    - Kuan-Han Lin - Carnegie Mellon University\n    - Christina Schenk - Basque Center for Applied Mathematics\n    - Michael Short - University of Surrey\n    - Jose Santiago Rodriguez - Purdue University\n    - David M. Thierry - Carnegie Mellon University\n    - Salvador García-Muñoz - Eli Lilly\n    - Lorenz T. Biegler - Carnegie Mellon University\n\nPlease cite\n------------\n - C. Schenk, M. Short, J.S. Rodriguez, D. Thierry, L.T. Biegler, S. García-Muñoz, W. Chen (2020)\nIntroducing KIPET: A novel open-source software package for kinetic parameter estimation from experimental datasets including spectra, Computers & Chemical Engineering, 134, 106716. https://doi.org/10.1016/j.compchemeng.2019.106716\n\n - M. Short, L.T. Biegler, S. García-Muñoz, W. Chen (2020)\nEstimating variances and kinetic parameters from spectra across multiple datasets using KIPET, Chemometrics and Intelligent Laboratory Systems, https://doi.org/10.1016/j.chemolab.2020.104012\n\n - M. Short, C. Schenk, D. Thierry, J.S. Rodriguez, L.T. Biegler, S. García-Muñoz (2019)\nKIPET–An Open-Source Kinetic Parameter Estimation Toolkit, Computer Aided Chemical Engineering, 47, 299-304.\n\n\n\n\n\n\n',
    'author': 'Kevin McBride',
    'author_email': 'kevin.w.mcbride.86@gmail.com',
    'maintainer': 'Kevin McBride',
    'maintainer_email': 'kevin.w.mcbride.86@gmail.com',
    'url': 'https://github.com/salvadorgarciamunoz/kipet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
KIPET 2020

This file contains the object functions used throughout Kipet modules in one
place.

"""
from pyomo.environ import (
    Objective,
    )

def get_objective(model, *args, **kwargs):
    
    objective_type = kwargs.get('objective_type', 'concentration')
    
    if objective_type == 'concentration':
        objective_expr = conc_objective(model, *args, **kwargs)

    return Objective(rule=objective_expr)


def conc_objective(model, *args, **kwargs):
    """
    
    Parameters
    ----------
    m : Pyomo ConcreteModel
        This is the current used in parameter fitting

    Returns
    -------
    obj : Objective function for Pyomo models
        This is the concentration based objective function

    """
    obj=0
  
    source = kwargs.get('source', 'concentration')
    
    if source == 'concentration':
    
        if model.mixture_components & model.measured_data:
            for index, values in model.Cm.items():
                obj += _concentration_term(model, index, var='Cm', **kwargs)
      
    elif source == 'spectra':
      
        if model.mixture_components:
            for index, values in model.C.items():
                obj += _concentration_term(model, index, var='C', **kwargs)
      
    return obj

def comp_objective(model, *args, **kwargs):
    """
    
    Parameters
    ----------
    m : Pyomo ConcreteModel
        This is the current used in parameter fitting

    Returns
    -------
    obj : Objective function for Pyomo models
        This is the concentration based objective function

    """
    obj=0
    
    if model.complementary_states & model.measured_data:
        for index, values in model.U.items():
            obj += _complementary_state_term(model, index)
        
    return obj

def spectra_objective(model, *args, **kwargs):
    """
    
    Parameters
    ----------
    m : Pyomo ConcreteModel
        This is the current used in parameter fitting

    Returns
    -------
    obj : Objective function for Pyomo models
        This is the concentration based objective function

    """
    obj=0
    # change this to items in the list (D or D_bar)
    # for t in model.meas_times:
    #     for l in model.meas_lambdas:
    
    for index, values in model.D.items():
        obj += _spectra_term(model, index)
        
    return obj

def absorption_objective(model, *args, **kwargs):
    """
    
    Parameters
    ----------
    m : Pyomo ConcreteModel
        This is the current used in parameter fitting

    Returns
    -------
    obj : Objective function for Pyomo models
        This is the concentration based objective function

    """
    sigma_device = kwargs.get('device_variance', 1)
    g_option = kwargs.get('g_option', None)
    with_d_vars = kwargs.get('with_d_vars', True)
    shared_spectra = kwargs.get('shared_spectra', True)
    list_components = kwargs.get('species_list', None)

    obj=0

    for index, values in model.D.items():
        obj += _spectral_term_MEE(model,
                                  index,
                                  sigma_device,
                                  g_option,
                                  shared_spectra,
                                  with_d_vars,
                                  list_components)
    return obj

# def calc_D_bar(model, D_bar_use, list_components):
    
#     if D_bar_use is False:
#         D_bar = model.D_bar
#     else:
#         D_bar = {}
#         if hasattr(model, '_abs_components'):
#             d_bar_list = model._abs_components
#             c_var = 'Cs'
#         else:
#             d_bar_list = list_components
#             c_var = 'C'    
            
#         if hasattr(model, 'huplc_absorbing') and hasattr(model, 'solid_spec_arg1'):
#             d_bar_list = [k for k in d_bar_list if k not in model.solid_spec_arg1]
                 
#         for t in model.meas_times:
#             for l in model.meas_lambdas:
#                 D_bar[t, l] = sum(getattr(model, c_var)[t, k] * model.S[l, k] for k in d_bar_list)

#     return D_bar
        
def _concentration_term(model, index, var='C', **kwargs):
    """
    
    Parameters
    ----------
    m : Pyomo ConcreteModel
        This is the current used in parameter fitting

    index : tuple
        This is the index of the model.C component

    Returns
    -------
    objective_concentration_term : Pyomo expression
        LS concentration term for objective

    """
    custom_sigma = kwargs.get('variance', None)
    
    if custom_sigma is None:
        variance = model.sigma[index[1]]
    else:
        variance = custom_sigma[index[1]]
        
    if variance is None:
        variance = 1
    
    objective_concentration_term = (getattr(model, var)[index] - model.Z[index]) ** 2  / variance
    
    return objective_concentration_term
    
def _complementary_state_term(model, index, **kwargs):
    """
    
    Parameters
    ----------
    m : Pyomo ConcreteModel
        This is the current used in parameter fitting

    index : tuple
        This is the index of the model.C component

    Returns
    -------
    objective_complementary_state_term : Pyomo expression
        LS complementary state term for objective

    """
    custom_sigma = kwargs.get('variance', None)
    
    if custom_sigma is None:
        variance = model.sigma[index[1]]
    else:
        variance = custom_sigma[index[1]]
    
    objective_complementary_state_term = (model.U[index] - model.X[index]) ** 2 / variance
    
    return objective_complementary_state_term

def _spectra_term(model, index, use_sigma=True):
    """
    
    Parameters
    ----------
    m : Pyomo ConcreteModel
        This is the current used in parameter fitting

    index : tuple
        This is the index of the model.C component

    Returns
    -------
    objective_complementary_state_term : Pyomo expression
        LS complementary state term for objective

    """
    objective_complementary_state_term = 0.5*(model.D[index] - model.D_bar[index]) ** 2
    
    if use_sigma:
        objective_complementary_state_term /= model.sigma['device']**2

    return objective_complementary_state_term
    
def _absorption_term(model, index, sigma_device=1, D_bar=None, g_options=None):
    
    print(index)
    print(model.D[index])
    print(D_bar[index])
    
    if g_options['unwanted_G'] or g_options['time_variant_G']:
        objective_absorption_term = (model.D[index] - D_bar[index] - model.qr[index[0]]*model.g[index[1]]) ** 2 / sigma_device
    elif g_options['time_invariant_G_no_decompose']:
        objective_absorption_term = (model.D[index] - D_bar[index] - model.g[index[1]]) ** 2 / sigma_device
    else:
        objective_absorption_term = (model.D[index] - D_bar[index]) ** 2 / sigma_device
        
    print(f'The index: {index}')
    print(objective_absorption_term.to_string())
    print('\n #### \n')

    return objective_absorption_term
    
def _spectral_term_MEE(model, index, sigma_device, g_option, shared_spectra, with_d_vars, list_components):
    
    t = index[0]
    l = index[1]
    if with_d_vars:
        base = model.D[t, l] - model.D_bar[t, l]
    else:
        D_bar = sum(model.C[t, k] * model.S[l, k] for k in list_components)
        base = model.D[t, l] - D_bar
        
    G_term = 0
    if g_option == 'time_variant_G':
        G_term -= model.qr[t]*model.g[l]
    elif g_option == 'time_invariant_G_decompose' and shared_spectra or g_option == 'time_invariant_G_no_decompose':
        G_term -= model.g[l]
   
    objective_spectral_term = (base + G_term)**2/sigma_device
    return objective_spectral_term




"""
ModelElement Blocks
"""
import kipet.model_components.ModelComponent as model_components
from kipet.top_level.variable_names import VariableNames

class ModelElementBlock():
    
    """Data abstraction for multiple ModelElement instances"""
    
    def __init__(self, class_name):
        
        attr_name = class_name.split('_')[-1] + 's'
        self.attr_class_set_name = attr_name
        setattr(self, attr_name, {})
        
        self.element_object_name = ''.join([term.capitalize() for term in class_name.split('_')])
        self.element_object = getattr(model_components, self.element_object_name)
        
        self._dict = getattr(self, self.attr_class_set_name)
        
    def __getitem__(self, value):
        
        return getattr(self, self.attr_class_set_name)[value]
    
    def __add__(self, other):
        
        return {**self._dict, **other._dict}
         
    def __str__(self):
        
        
        format_string = "{:<10}{:<15}{:<15}\n"
        param_str = f'{self.element_object_name}:\n'
        param_str += format_string.format(*['Name', 'Value', 'Units'])
        
        for elem in self._dict.values():
    
            elem_name = elem.name
            elem_value = 'None' if elem.value is None else elem.value
            elem_units = 'None' if elem.units is None or elem.units  else elem.units

            param_str += format_string.format(elem_name, elem_value, elem_units)
        
        return param_str

    def __repr__(self):
        return self.__str__()
        # return self.element_object_name

    def __iter__(self):
        for param, data in getattr(self, self.attr_class_set_name).items():
            yield data
            
    def __len__(self):
        return len(getattr(self, self.attr_class_set_name))
    
    def __contains__(self, key):
        return key in self._dict
    

    def add_element_list(self, elem_list):
        """Handles lists of parameters or single parameters added to the model
       
        """
        for elem in elem_list:
            self.add_element(*elem)        
        
        return None
    
    def add_element(self, *args, **kwargs):
        
        """
            
        """
        name = args[0]
        element = self.element_object(name,
                                      **kwargs,
                                      )
            
        getattr(self, self.attr_class_set_name)[element.name] = element
        
    def as_dict(self, attr):
        
        return_dict = {}
        for param, obj in self._dict.items():
            return_dict[param] = getattr(obj, attr)
        return return_dict
        
    def update(self, attr, dict_data):
        
        for elem, new_data in dict_data.items():
            if elem in self._dict:
                setattr(self._dict[elem], attr, new_data)

        return None
    
    def get_match(self, attr, query):
        
        query_list = []
        for elem, obj in self._dict.items():
            if getattr(obj, attr) == query:
                query_list.append(elem)
                
        return query_list
    
    
    @property 
    def names(self):
        return [elem for elem in getattr(self, self.attr_class_set_name)]
    
    
class ConstantBlock(ModelElementBlock):
    
    def __init__(self, *args, **kwargs):
        super().__init__(class_name='model_constant')
        
class AlgebraicBlock(ModelElementBlock):
    
    __var = VariableNames()
    
    def __init__(self, *args, **kwargs):
        super().__init__(class_name='model_algebraic')
       
    @property
    def fixed(self):
        
        fix_from_traj = []
        for alg in self.algebraics.values():
            if alg.data is not None:
                fix_from_traj.append([self.__var.algebraic, alg.name, alg.data])
        
        return fix_from_traj
    
    @property
    def steps(self):
        
        steps = {}
        for alg in self.algebraics.values():
            if alg.step is not None:
                steps[alg.name] = alg
        
        return steps
        
class ComponentBlock(ModelElementBlock):
    
    def __init__(self, *args, **kwargs):
        super().__init__(class_name='model_component')
    
    
    def var_variances(self):
        
        sigma_dict = {}
        
        for component in self.components.values():       
            if component.state == 'trajectory':
                continue       
            sigma_dict[component.name] = component.variance
            
        return sigma_dict
        
    def component_set(self, category):
        
        component_set = []
        for component in self.components.values():
            if component.state == category:
                component_set.append(component.name)
                
        return component_set
    
    @property
    def variances(self):
        return {comp.name: comp.variance for comp in self.components.values()}
    
    @property
    def init_values(self):
        return {comp.name: comp.value for comp in self.components.values()}
    
    @property
    def known_values(self):
        return {comp.name: comp.known for comp in self.components.values()}
        
    @property
    def names(self):
        return [comp.name for comp in self.components.values()]
    
    @property
    def has_all_variances(self):
    
        all_component_variances = True
        for comp in self.components.values():
            if comp.variance is None:
                all_component_variances = False
            if not all_component_variances:
                break
        return all_component_variances
    
class StateBlock(ModelElementBlock):
    
    def __init__(self, *args, **kwargs):
        super().__init__(class_name='model_state')
    
    
    def var_variances(self):
        
        sigma_dict = {}
        
        for component in self.states.values():       
            if component.state == 'trajectory':
                continue       
            sigma_dict[component.name] = component.variance
            
        return sigma_dict
        
    def component_set(self, category):
        
        component_set = []
        for component in self.states.values():
            if component.state == category:
                component_set.append(component.name)
                
        return component_set
    
    @property
    def variances(self):
        return {comp.name: comp.variance for comp in self.states.values()}
    
    @property
    def init_values(self):
        return {comp.name: comp.value for comp in self.states.values()}
    
    @property
    def known_values(self):
        return {comp.name: comp.known for comp in self.states.values()}
        
    @property
    def names(self):
        return [comp.name for comp in self.states.values()]
    
    @property
    def has_all_variances(self):
    
        all_component_variances = True
        for comp in self.components.values():
            if comp.variance is None:
                all_component_variances = False
            if not all_component_variances:
                break
        return all_component_variances
    
    
class ParameterBlock(ModelElementBlock):
    
    def __init__(self, *args, **kwargs):
        super().__init__(class_name='model_parameter')
    
    @property
    def lb(self):
        """Lower bound property"""
        return self.bounds[0]

    @property
    def ub(self):
        """Upper bound property"""
        return self.bounds[1]


"""
Component expression handling and units model development
"""
# Standard library imports

# Third party imports
from pyomo.core.base.PyomoModel import ConcreteModel
from pyomo.core.base.var import Var
from pyomo.environ import units as u

# KIPET library imports
from kipet.post_model_build.pyomo_model_tools import get_index_sets
from kipet.common.VisitorClasses import ReplacementVisitor
from kipet.post_model_build.replacement import _update_expression
from kipet.top_level.variable_names import VariableNames

string_replace_dict = {
    ' ': '_',
    '-': 'n',
    '+': 'p',
    '.': '_',
    }

class ModalVar():
        
    def __init__(self, name, comp, index, model):
        
        self.name = name
        self.comp = comp
        self.index = index
        self.model = model
    
class Comp():
    
    var = VariableNames()
    
    def __init__(self, model, constant_info=None):
        
        self._model = model
        self._model_vars = self.var.model_vars
        self._rate_vars = self.var.rate_vars
        self.constant_info = constant_info
        self.var_dict = {}
        self.var_unit_dict = {}
        self.assign_vars()
        self.assign_rate_vars()
        
    def assign_vars(self):
        """Digs through and assigns the variables as top-level attributes
        
        """
        list_of_vars = self._model_vars
        
        for mv in list_of_vars:
            if hasattr(self._model, mv):
                mv_obj = getattr(self._model, mv)
                index_sets = get_index_sets(mv_obj)
                
                dim = len(index_sets)
                if dim > 1:
                    indx = 1
                else:
                    indx = 0
                
                comp_set = list(index_sets[indx].keys())
                for comp in comp_set:
                    
                    comp = str(comp)
                    if isinstance(comp, str):
                        if isinstance(comp[0], int):
                            comp.insert(0, 'y')
                        
                        comp_name = comp
                        for k, v in string_replace_dict.items():
                            comp_name.replace(k, v)
                    
                    self.var_dict[comp_name] = ModalVar(comp_name, comp_name, mv, self._model)
                    
                    if dim > 1:
                        setattr(self, comp_name, mv_obj[0, comp])
                    else:
                        setattr(self, comp_name, mv_obj[comp])

    def assign_rate_vars(self):

        for mv in self._rate_vars:
            if hasattr(self._model, mv):
                
                mv_obj = getattr(self._model, mv)
                index_sets = get_index_sets(mv_obj)
                comp_set = list(index_sets[-1].keys())
                dim = len(index_sets)
                
                for comp in comp_set:
                    
                    if isinstance(comp, str):
                        comp_name = comp
                        comp_name.replace(' ', '_')
                    elif isinstance(comp, int):
                        comp_name = f'y{comp}'
                    
                    comp_name = f'd{comp_name}dt'
                    self.var_dict[comp_name] = ModalVar(comp_name, str(comp), mv, self._model)
                    if dim > 1:
                        setattr(self, comp_name, mv_obj[0, comp])
                    else:
                        setattr(self, comp_name, mv_obj[comp])  

    def _id(self, comp):
        
        id_ = id(getattr(self, comp))
        print(id_)
        return id_
    
    @property
    def get_var_list(self):
        
        return list(self.var_dict.keys())
    
    
# class Comp_Check():
    
#     var = VariableNames()
    
#     def __init__(self, model, param_list):
        
#         self._model = model
#         self._param_list = param_list
#         self.var_dict = {}
#         self.assign_params_for_units_check()
        
#     def assign_params_for_units_check(self):
#         """Digs through and assigns the variables as top-level attributes
        
#         """
#         m = self._model
#         for var in self._param_list:
#             if hasattr(m, var):
#                 var_obj = getattr(m, var)
#                 self.var_dict[var] = var_obj
#                 setattr(self, var, var_obj)
                
# def get_unit_model(element_dict, set_up_model):
#     """Takes a ReactionModel instance and returns a Comp_Check object
#     representing the model's varialbes with units
#     """            
#     unit_model = ConcreteModel()
#     new_params = set()
#     for elem, comp_block in element_dict.items():
#         if len(comp_block) > 0:
#             for comp in comp_block:
                
#                 new_params.add(comp.name)
#                 m_units = comp.units
#                 if m_units is None:
#                     m_units = str(1)
#                 else:
#                     m_units = getattr(u, str(m_units))
                    
#                 print(m_units)
#                 print(type(m_units))
#                 setattr(unit_model, comp.name, Var(initialize=1, units=m_units))
        
#     c_units = Comp_Check(unit_model, list(new_params))
    
#     return c_units